/***************************************************************************
 *   Copyright (C) 2008 by Georg Hennig                                    *
 *   georg.hennig@web.de                                                   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/


#include "kfile_krawimageinfo.h"

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <klocale.h>
#include <kgenericfactory.h>
#include <kdebug.h>
#include <ktempfile.h>
#include <kimageio.h>
#include <qfile.h>
#include <qimage.h>
#include <qwmatrix.h>
#include <cstdio>

#include <libkexiv2/kexiv2.h>

using namespace KExiv2Iface;

typedef KGenericFactory<KRawImageInfo> RawFactory;

K_EXPORT_COMPONENT_FACTORY( kfile_raw, RawFactory( "kfile_raw" ) )

#ifndef KDE_EXPORT
# define KDE_EXPORT
#endif

KRawImageInfo::KRawImageInfo( QObject *parent, const char *name, const QStringList &args )
	: KFilePlugin( parent, name, args )
{
  kdDebug(7034) << "KCameraRawPlugin c'tor" << endl;

  //
  // define all possible meta info items
  //
  KFileMimeTypeInfo *info = addMimeTypeInfo("image/x-raw");
  KFileMimeTypeInfo::GroupInfo *exifGroup = addGroupInfo( info, "Info",
						      i18n("Image Info") );
  KFileMimeTypeInfo::ItemInfo* item;

  item = addItemInfo( exifGroup, "Manufacturer", i18n("Camera Manufacturer"),
                      QVariant::String );

  item = addItemInfo( exifGroup, "Model", i18n("Camera Model"),
                      QVariant::String );

  item = addItemInfo( exifGroup, "Date/time", i18n("Date/Time"),
                      QVariant::DateTime );

  item = addItemInfo( exifGroup, "CreationDate", i18n("Creation Date"),
                      QVariant::Date );

  item = addItemInfo( exifGroup, "CreationTime", i18n("Creation Time"),
                      QVariant::Time );

  item = addItemInfo( exifGroup, "Dimensions", i18n("Dimensions"),
                      QVariant::Size );
  setHint( item, KFileMimeTypeInfo::Size );
  setUnit( item, KFileMimeTypeInfo::Pixels );

  item = addItemInfo( exifGroup, "Orientation", i18n("Orientation"),
                      QVariant::Int );

  item = addItemInfo( exifGroup, "ColorMode", i18n("Color Mode"),
                      QVariant::String );

  item = addItemInfo( exifGroup, "Flash used", i18n("Flash Used"),
                      QVariant::String );
  item = addItemInfo( exifGroup, "Focal length", i18n("Focal Length"),
                      QVariant::String );
  setUnit( item, KFileMimeTypeInfo::Millimeters );

  item = addItemInfo( exifGroup, "35mm equivalent", i18n("35mm Equivalent"),
                      QVariant::Int );
  setUnit( item, KFileMimeTypeInfo::Millimeters );

  item = addItemInfo( exifGroup, "CCD width", i18n("CCD Width"),
                      QVariant::String );
  setUnit( item, KFileMimeTypeInfo::Millimeters );

  item = addItemInfo( exifGroup, "Exposure time", i18n("Exposure Time"),
                      QVariant::String );
  setHint( item, KFileMimeTypeInfo::Seconds );

  item = addItemInfo( exifGroup, "Aperture", i18n("Aperture"),
                      QVariant::String );

  item = addItemInfo( exifGroup, "Focus dist.", i18n("Focus Dist."),
                      QVariant::String );

  item = addItemInfo( exifGroup, "Exposure bias", i18n("Exposure Bias"),
                      QVariant::String );

  item = addItemInfo( exifGroup, "Whitebalance", i18n("Whitebalance"),
                      QVariant::String );

  item = addItemInfo( exifGroup, "Metering mode", i18n("Metering Mode"),
                      QVariant::String );

  item = addItemInfo( exifGroup, "Exposure", i18n("Exposure"),
                      QVariant::String );

  item = addItemInfo( exifGroup, "ISO equiv.", i18n("ISO Equiv."),
                      QVariant::String );

  item = addItemInfo( exifGroup, "User comment", i18n("User Comment"),
                      QVariant::String );
  setHint(item,  KFileMimeTypeInfo::Description);

  item = addItemInfo( exifGroup, "Thumbnail", i18n("Thumbnail"),
                      QVariant::Image );
  setHint( item, KFileMimeTypeInfo::Thumbnail );
}

bool KRawImageInfo::readInfo( KFileMetaInfo& info, uint what )
{
	kdDebug(7034) << "KCameraRawPlugin::readInfo()" << endl;

	const QString path( info.path() );
	if ( path.isEmpty() ) // remote file
		return false;

	KFileMetaInfoGroup group = appendGroup( info, "Info" );

	KExiv2 my_kexiv2( path );

	long exiv_numerator = 0;
	long exiv_denominator = 0;
	long exiv_long1 = 0;
	long exiv_long2 = 0;

	if ( !my_kexiv2.getExifTagString( "Exif.Image.Make" ).isEmpty() )
		appendItem( group, "Manufacturer", my_kexiv2.getExifTagString( "Exif.Image.Make" ) );

	if ( !my_kexiv2.getExifTagString( "Exif.Image.Model" ).isEmpty() )
		appendItem( group, "Model", my_kexiv2.getExifTagString( "Exif.Image.Model" ) );

	if ( !my_kexiv2.getExifTagString( "Exif.Image.DateTime" ).isEmpty() )
	{
		QDateTime dt = parseDateTime( my_kexiv2.getExifTagString( "Exif.Image.DateTime" ).stripWhiteSpace() );
		if ( dt.isValid() )
		{
			appendItem( group, "Date/time", dt );
			appendItem( group, "CreationDate", dt.date() );
			appendItem( group, "CreationTime", dt.time() );
		}
	}

	if ( my_kexiv2.getExifTagLong( "Exif.Photo.PixelXDimension", exiv_long1 ) && my_kexiv2.getExifTagLong( "Exif.Photo.PixelYDimension", exiv_long2 ) )
	{
		QSize size = QSize( exiv_long1, exiv_long2 );
		appendItem( group, "Dimensions", size );
	}

	if ( my_kexiv2.getExifTagLong( "Exif.Image.Orientation", exiv_long1 ) )
	{
		appendItem( group, "Orientation", (int)exiv_long1 );
	}

	if ( !my_kexiv2.getExifTagString( "Exif.Photo.ColorSpace" ).isEmpty() )
	{
		appendItem( group, "ColorMode", my_kexiv2.getExifTagString( "Exif.Photo.ColorSpace" ) );
	}

	if ( my_kexiv2.getExifTagLong( "Exif.Photo.Flash", exiv_long1 ) )
	{
		QString flash = i18n( "(unknown)" );
		switch ( exiv_long1 )
		{
			case 0:
				flash = i18n( "No" );
				break;
			case 1:
			case 5:
			case 7:
				flash = i18n( "Fired" );
				break;
			case 9:
			case 13:
			case 15:
				flash = i18n( "Fill Fired" );
				break;
			case 16:
				flash = i18n( "Off" );
				break;
			case 24:
				flash = i18n( "Auto Off" );
				break;
			case 25:
			case 29:
			case 31:
				flash = i18n( "Auto Fired" );
				break;
			case 32:
				flash = i18n( "Not Available" );
				break;
			default:
				break;
		}

		appendItem( group, "Flash used", flash );
	}

	double ccd_width = 0.;
	if ( my_kexiv2.getExifTagLong( "Exif.Photo.FocalPlaneResolutionUnit", exiv_long1 ) &&
		my_kexiv2.getExifTagRational( "Exif.Photo.FocalPlaneXResolution", exiv_numerator, exiv_denominator ) &&
		my_kexiv2.getExifTagLong( "Exif.Photo.PixelXDimension", exiv_long2 ) )
	{
		ccd_width = (double)(exiv_long2*exiv_denominator)/exiv_numerator;
		switch ( exiv_long1 )
		{
			case 1: ccd_width *= 25.4; break; // inch
			case 2:
				// According to the information I was using, 2 means meters.
				// But looking at the Cannon powershot's files, inches is the only
				// sensible value.
				ccd_width *= 25.4;
				break;
			case 3: ccd_width *= 10;   break;  // centimete
			case 4: break;  // milimeter
			case 5: ccd_width *= .001; break;  // micrometer
		}

		appendItem( group, "CCD width", QString().sprintf( "%4.2f", ccd_width ) );
	}

	if ( my_kexiv2.getExifTagRational( "Exif.Photo.FocalLength", exiv_numerator, exiv_denominator ) )
	{
		if ( exiv_numerator > 0 )
		{
			appendItem( group, "Focal length", QString().sprintf( "%4.1f", (double)(exiv_numerator)/exiv_denominator ) );
		}

		if ( my_kexiv2.getExifTagLong( "Exif.Photo.FocalLengthIn35mmFilm", exiv_long1 ) )
		{
			appendItem( group, "35mm equivalent", (int)exiv_long1 );
		}
		else if ( ccd_width > 0. && exiv_numerator > 0 )
		{
			appendItem( group, "35mm equivalent", (int)((double)(exiv_numerator)/(exiv_denominator*ccd_width)*35 + 0.5) );
		}
	}

	if ( my_kexiv2.getExifTagRational( "Exif.Photo.ExposureTime", exiv_numerator, exiv_denominator ) )
	{
		QString exposure_time;

		if ( exiv_numerator < 0 || exiv_denominator < 0 || exiv_numerator > LONG_MAX || exiv_denominator > LONG_MAX )
		{
			if ( my_kexiv2.getExifTagLong( "Exif.CanonSi.0x0018", exiv_long1 ) ) // Bulb exposure time
			{
				if ( exiv_long1 >= 10 ) exposure_time = QString().sprintf( "%6.3f", (double)(exiv_long1/10) );
				else exposure_time = i18n( "Unknown" );
			}
			else
			{
				exposure_time = i18n( "Unknown" );
			}
		}
		else
		{
			exposure_time = QString().sprintf( "%6.3f", (double)(exiv_numerator)/exiv_denominator );
			if ( exiv_numerator % exiv_denominator != 0 && (double)(exiv_numerator)/exiv_denominator < 1. )
				exposure_time += QString().sprintf( " (1/%d)", (int)exiv_denominator );
		}

		appendItem( group, "Exposure time", exposure_time );
	}

	if ( my_kexiv2.getExifTagRational( "Exif.Photo.FNumber", exiv_numerator, exiv_denominator ) )
	{
		if ( exiv_numerator > 0 )
		{
			appendItem( group, "Aperture", QString().sprintf( "f/%3.1f", (double)(exiv_numerator)/exiv_denominator ) );
		}
	}

	if ( my_kexiv2.getExifTagRational( "Exif.Photo.SubjectDistance", exiv_numerator, exiv_denominator ) )
	{
		QString subject_distance;
		if ( exiv_numerator == 0 ) subject_distance = i18n( "Unknown" );
		else if ( exiv_denominator <= 0 || exiv_numerator < 0 ) i18n( "Infinite" );
		else subject_distance = QString().sprintf( "%5.2fm", (double)(exiv_numerator)/exiv_denominator );

		appendItem( group, "Focus dist.", subject_distance );
	}


	if ( my_kexiv2.getExifTagRational( "Exif.Photo.ExposureBiasValue", exiv_numerator, exiv_denominator ) )
	{
		appendItem( group, "Exposure bias", QString().sprintf( "%4.2f", (double)(exiv_numerator)/exiv_denominator ) );
	}

	if ( my_kexiv2.getExifTagLong( "Exif.Photo.WhiteBalance", exiv_long1 ) )
	{
		QString white_balance;
		switch ( exiv_long1 )
		{
			case 0: white_balance=i18n("Unknown"); break;
			case 1: white_balance=i18n("Daylight"); break;
			case 2: white_balance=i18n("Fluorescent"); break;
			case 3: white_balance=i18n("Tungsten"); break;
			case 17: white_balance=i18n("Standard light A"); break;
			case 18: white_balance=i18n("Standard light B"); break;
			case 19: white_balance=i18n("Standard light C"); break;
			case 20: white_balance=i18n("D55"); break;
			case 21: white_balance=i18n("D65"); break;
			case 22: white_balance=i18n("D75"); break;
			case 255: white_balance=i18n("Other"); break;
			default: //23 to 254 = reserved
				white_balance=i18n("Unknown");
		}
  	appendItem( group, "Whitebalance", white_balance );
	}

	if ( my_kexiv2.getExifTagLong( "Exif.Photo.MeteringMode", exiv_long1 ) )
	{
		QString metering_mode;
		switch ( exiv_long1 )
		{
			case 0: metering_mode=i18n("Unknown"); break;
			case 1: metering_mode=i18n("Average"); break;
			case 2: metering_mode=i18n("Center weighted average"); break;
			case 3: metering_mode=i18n("Spot"); break;
			case 4: metering_mode=i18n("MultiSpot"); break;
			case 5: metering_mode=i18n("Pattern"); break;
			case 6: metering_mode=i18n("Partial"); break;
			case 255: metering_mode=i18n("Other"); break;
			default: // 7 to 254 = reserved
				metering_mode=i18n("Unknown");
		}
  	appendItem( group, "Metering mode", metering_mode );
	}

	if ( my_kexiv2.getExifTagLong( "Exif.Photo.ExposureMode", exiv_long1 ) )
	{
		QString exposure_mode;
		switch( exiv_long1 )
		{
			case 0: exposure_mode=i18n("Not defined"); break;
			case 1: exposure_mode=i18n("Manual"); break;
			case 2: exposure_mode=i18n("Normal program"); break;
			case 3: exposure_mode=i18n("Aperture priority"); break;
			case 4: exposure_mode=i18n("Shutter priority"); break;
			case 5: exposure_mode=i18n("Creative program\n(biased toward fast shutter speed)"); break;
			case 6: exposure_mode=i18n("Action program\n(biased toward fast shutter speed)"); break;
			case 7: exposure_mode=i18n("Portrait mode\n(for closeup photos with the background out of focus)"); break;
			case 8: exposure_mode=i18n("Landscape mode\n(for landscape photos with the background in focus)"); break;
			default: // 9 to 255 = reserved
				exposure_mode=i18n("Unknown");
		}
		appendItem( group, "Exposure", exposure_mode );
	}

	if ( my_kexiv2.getExifTagLong( "Exif.Photo.ISOSpeedRatings", exiv_long1 ) )
	{
		appendItem( group, "ISO equiv.", QString().sprintf( "%2d", (int)exiv_long1 ) );
	}

	if ( !my_kexiv2.getExifTagString( "Exif.Photo.UserComment" ).isEmpty() )
	{
		appendItem( group, "User comment", my_kexiv2.getExifTagString( "Exif.Photo.UserComment" ) );
	}

	if ( what & KFileMetaInfo::Thumbnail )
	{
		QImage img = my_kexiv2.getExifThumbnail( false );

		if( !img.isNull() )
		{

		long orientation;
		if ( my_kexiv2.getExifTagLong( "Exif.Image.Orientation", orientation ) )
		{
				QWMatrix M;
				QWMatrix flip= QWMatrix(-1,0,0,1,0,0);
				switch( orientation )
				{
					case 2: M = flip; break;
					case 4: M = flip;
					case 3: M.rotate(180); break;
					case 5: M = flip;
					case 6: M.rotate(90); break;
					case 7: M = flip;
					case 8: M.rotate(270); break;
				}
				img = img.xForm(M);
			}

			if (img.depth() != 32) img = img.convertDepth(32);

			appendItem( group, "Thumbnail", img );
		}
	}

	return true;
}

// format of the string is:
// YYYY:MM:DD HH:MM:SS
QDateTime KRawImageInfo::parseDateTime( const QString& string )
{
    QDateTime dt;
    if ( string.length() != 19 )
        return dt;

    QString year    = string.left( 4 );
    QString month   = string.mid( 5, 2 );
    QString day     = string.mid( 8, 2 );
    QString hour    = string.mid( 11, 2 );
    QString minute  = string.mid( 14, 2 );
    QString seconds = string.mid( 17, 2 );

    bool ok;
    bool allOk = true;
    int y  = year.toInt( &ok );
    allOk &= ok;

    int mo = month.toInt( &ok );
    allOk &= ok;

    int d  = day.toInt( &ok );
    allOk &= ok;

    int h  = hour.toInt( &ok );
    allOk &= ok;

    int mi = minute.toInt( &ok );
    allOk &= ok;

    int s  = seconds.toInt( &ok );
    allOk &= ok;

    if ( allOk ) {
        dt.setDate( QDate( y, mo, d ) );
        dt.setTime( QTime( h, mi, s ) );
    }

    return dt;
}

